<?php
/**
 * Luxor_Lang:: defines an API for the different programming languages Luxor
 * is able to parse.
 *
 * $Horde: luxor/lib/Lang.php,v 1.7 2004/05/20 15:56:18 jan Exp $
 *
 * @author  Jan Schneider <jan@horde.org>
 * @version $Revision: 1.7 $
 * @since   Luxor 0.1
 * @package Luxor
 */
class Luxor_Lang {

    /**
     * Attempts to return a concrete Luxor_Lang instance based on $driver.
     *
     * @param string    $driver     The type of concrete Luxor_Lang subclass
     *                              to return.  The is based on the repository
     *                              driver ($driver).  The code is dynamically
     *                              included.
     * @param array     $params     (optional) A hash containing any additional
     *                              configuration or connection parameters a
     *                              subclass might need.
     *
     * @return mixed    The newly created concrete Luxor_Lang instance, or
     *                  false on an error.
     */
    function &factory($driver, $params = array())
    {
        $driver = basename($driver);
        require_once dirname(__FILE__) . '/Lang/' . $driver . '.php';
        $class = 'Luxor_Lang_' . $driver;
        if (class_exists($class)) {
            return $ret = &new $class($params);
        } else {
            return false;
        }
    }

    /**
     * Attempts to return a reference to a concrete Luxor_Lang instance
     * based on $driver.  It wil only create a new instance if no
     * Luxor_Lang instance with the same parameters currently exists.
     *
     * This should be used if multiple language parsers are required.
     *
     * This method must be invoked as: $var = &Luxor_Lang::singleton()
     *
     * @param string    $driver     The type of concrete Luxor_Lang subclass
     *                              to return.  The is based on the repository
     *                              driver ($driver).  The code is dynamically
     *                              included.
     * @param array     $params     (optional) A hash containing any additional
     *                              configuration or connection parameters a
     *                              subclass might need.
     *
     * @return mixed    The created concrete Luxor_Lang instance, or false
     *                  on error.
     */
    function &singleton($driver, $params = array())
    {
        static $instances;

        if (!isset($instances)) {
            $instances = array();
        }

        $signature = serialize(array($driver, $params));
        if (!isset($instances[$signature])) {
            $instances[$signature] = &Luxor_Lang::factory($driver, $params);
        }

        return $instances[$signature];
    }

    /**
     * Attempts to determine a files programming language and returns
     * a parser instance for this language.
     *
     * @param Luxor_Files $files  An instance of Luxor_Files to use for file
     *                            operations and path name resolution.
     * @param string $pathname    The path name of the file to create a
     *                            parser for.
     *
     * @return mixed    The created concrete Luxor_Lang instance, or false
     *                  on error.
     */
    function &builder($files, $pathname)
    {
        global $languages;

        require_once LUXOR_BASE . '/config/languages.php';

        /* First, check the 'filetype' hash for a matching file extension. */
        foreach ($languages['filetype'] as $type) {
            if (preg_match('/' . $type[1] . '/', $pathname)) {
                return Luxor_Lang::singleton($type[2], $type);
            }
        }

        /* Next, try to detect the shebang line. */
        $fh = $files->getFileHandle($pathname);
        if (!$fh || is_a($fh, 'PEAR_Error')) {
            return $fh;
        }
        $line = fgets($fh);
        if (!preg_match('/^\#!\s*(\S+)/s', $line, $match)) {
            return false;
        }
        if (array_key_exists($match[1], $languages['interpreters'])) {
            $lang = $languages['filetype'][$languages['interpreters'][$match[1]]];
            return Luxor_Lang::singleton($lang[2], $lang);
        }
        return false;
    }

    function processInclude($frag, $dir)
    {
        $frag = preg_replace('/([\'"])(.*?)([\'"])/e',
            'stripslashes(\'$1\') . Luxor::incRef(\'$2\', "fixed", \'$2\', $dir) . stripslashes(\'$3\')', $frag);
        $frag = preg_replace('/(\\0<)(.*?)(\\0>)/e',
            'stripslashes(\'$1\') . Luxor::incRef(\'$2\', "fixed", \'$2\') . stripslashes(\'$3\')', $frag);
        return $frag;
    }

}
